import array
import numbers
from cpython cimport array
from libc.stdint cimport uint8_t

cimport _heatshrink


MIN_WINDOW_SZ2 = _heatshrink.HEATSHRINK_MIN_WINDOW_BITS
MAX_WINDOW_SZ2 = _heatshrink.HEATSHRINK_MAX_WINDOW_BITS
DEFAULT_WINDOW_SZ2 = 11

MIN_LOOKAHEAD_SZ2 = _heatshrink.HEATSHRINK_MIN_LOOKAHEAD_BITS
DEFAULT_LOOKAHEAD_SZ2 = 4

DEFAULT_INPUT_BUFFER_SIZE = 2048


def _validate_bounds(val, name, min=None, max=None):
    """
    Ensure that `val` is larger than `min` and smaller than `max`.

    Throws `ValueError` if constraints are not met or
    if both `min` and `max` are None.
    Throws `TypeError` if `val` is not a number.
    """
    if min is None and max is None:
        raise ValueError("Expecting either a min or max parameter")

    if not isinstance(val, numbers.Number):
        msg = 'Expected number, got {}'
        raise TypeError(msg.format(val.__class__.__name__))

    if min and val < min:
        msg = "{} must be > {}".format(name, min)
    elif max and val > max:
        msg = "{} must be < {}".format(name, max)
    else:
        msg = ''

    if msg:
        raise ValueError(msg)
    return val


cdef class Writer:
    """Thin wrapper around heatshrink_encoder"""
    cdef _heatshrink.heatshrink_encoder *_hse

    def __cinit__(self, **kwargs):
        window_sz2 = kwargs.get('window_sz2', DEFAULT_WINDOW_SZ2)
        lookahead_sz2 = kwargs.get('lookahead_sz2', DEFAULT_LOOKAHEAD_SZ2)

        _validate_bounds(window_sz2, name='window_sz2',
                         min=MIN_WINDOW_SZ2, max=MAX_WINDOW_SZ2)
        _validate_bounds(lookahead_sz2, name='lookahead_sz2',
                         min=MIN_LOOKAHEAD_SZ2, max=window_sz2)

        self._hse = _heatshrink.heatshrink_encoder_alloc(window_sz2, lookahead_sz2)
        if self._hse is NULL:
            raise MemoryError('Failed to allocate encoder.')

    def __dealloc__(self):
        if self._hse is not NULL:
            _heatshrink.heatshrink_encoder_free(self._hse)

    @property
    def max_output_size(self):
        return 1 << self._hse.window_sz2

    def sink(self, array.array in_buf, size_t offset=0):
        """
        Sink input in to the encoder with an optional N byte `offset`.
        """
        cdef:
            size_t sink_size
            _heatshrink.HSE_sink_res res

            size_t in_buf_size = len(in_buf)

        res = _heatshrink.heatshrink_encoder_sink(
            self._hse,
            &in_buf.data.as_uchars[offset],
            in_buf_size - offset,
            &sink_size
        )
        return res, sink_size

    def poll(self, array.array out_buf):
        """Poll data from state machine in to an array.

        Assumes that the passed in array is large enough to
        contain all data from the state machine.
        """
        cdef:
            size_t poll_size
            _heatshrink.HSE_poll_res res

            size_t out_buf_size = len(out_buf)

        res = _heatshrink.heatshrink_encoder_poll(
            self._hse,
            out_buf.data.as_uchars,
            out_buf_size,
            &poll_size
        )
        return res, poll_size

    def is_poll_empty(self, _heatshrink.HSE_poll_res res):
        return res == _heatshrink.HSER_POLL_EMPTY

    def finish(self):
        """Notifies the encoder that the input stream is finished."""
        return _heatshrink.heatshrink_encoder_finish(self._hse)

    def is_finished(self, _heatshrink.HSE_finish_res res):
        return res == _heatshrink.HSER_FINISH_DONE


cdef class Reader:
    """Thin wrapper around heatshrink_decoder"""
    cdef _heatshrink.heatshrink_decoder *_hsd

    def __cinit__(self, **kwargs):
        input_buffer_size = kwargs.get('input_buffer_size',
                                       DEFAULT_INPUT_BUFFER_SIZE)
        window_sz2 = kwargs.get('window_sz2', DEFAULT_WINDOW_SZ2)
        lookahead_sz2 = kwargs.get('lookahead_sz2', DEFAULT_LOOKAHEAD_SZ2)

        _validate_bounds(input_buffer_size, name='input_buffer_size', min=0)
        _validate_bounds(window_sz2, name='window_sz2',
                         min=MIN_WINDOW_SZ2, max=MAX_WINDOW_SZ2)
        _validate_bounds(lookahead_sz2, name='lookahead_sz2',
                         min=MIN_LOOKAHEAD_SZ2, max=window_sz2)

        self._hsd = _heatshrink.heatshrink_decoder_alloc(
            input_buffer_size, window_sz2, lookahead_sz2)
        if self._hsd is NULL:
            raise MemoryError('Failed to allocate decoder.')

    def __dealloc__(self):
        if self._hsd is not NULL:
            _heatshrink.heatshrink_decoder_free(self._hsd)

    @property
    def max_output_size(self):
        return 1 << self._hsd.window_sz2

    def sink(self, array.array in_buf, size_t offset=0):
        """
        Sink input in to the encoder with an optional N byte `offset`.
        """
        cdef:
            size_t sink_size
            _heatshrink.HSD_sink_res res

            size_t in_buf_size = len(in_buf)

        res = _heatshrink.heatshrink_decoder_sink(
            self._hsd,
            &in_buf.data.as_uchars[offset],
            in_buf_size - offset,
            &sink_size
        )
        return res, sink_size

    def poll(self, array.array out_buf):
        cdef:
            size_t poll_size
            _heatshrink.HSD_poll_res res

            size_t out_buf_size = len(out_buf)

        res = _heatshrink.heatshrink_decoder_poll(
            self._hsd,
            out_buf.data.as_uchars,
            out_buf_size,
            &poll_size
        )
        return res, poll_size

    def is_poll_empty(self, _heatshrink.HSD_poll_res res):
        return res == _heatshrink.HSDR_POLL_EMPTY

    def finish(self):
        """Notifies the encoder that the input stream is finished."""
        return _heatshrink.heatshrink_decoder_finish(self._hsd)

    def is_finished(self, _heatshrink.HSD_finish_res res):
        return res == _heatshrink.HSDR_FINISH_DONE


class Encoder(object):
    """High level interface to the Heatshrink encoders/decoders."""
    def __init__(self, encoder):
        self._encoder = encoder
        self._finished = False

    def _check_not_finished(self):
        """Throws an exception if the encoder has been closed."""
        if self._finished:
            msg = 'Attempted to perform operation on a closed encoder.'
            # TODO: ValueError isn't the right exception for this
            raise ValueError(msg)

    def _drain(self):
        """Empty data from the encoder state machine."""
        cdef:
            array.array out_buf = array.array('B', [])

        while True:
            # Resize to decent length
            array.resize(out_buf, self._encoder.max_output_size)

            res, poll_size = self._encoder.poll(out_buf)
            if res < 0:
                raise RuntimeError('Encoder poll failed.')

            # Drop unused elements
            array.resize(out_buf, poll_size)

            yield out_buf

            # Done polling
            if self._encoder.is_poll_empty(res):
                raise StopIteration

    def fill(self, buf):
        """Fill the encoder state machine with a buffer."""
        self._check_not_finished()

        if isinstance(buf, (unicode, memoryview)):
            msg = "Cannot fill encoder with type '{.__name__}'"
            raise TypeError(msg.format(buf.__class__))

        # Convert input to a byte representation
        cdef array.array in_buf  = array.array('B', buf)
        cdef array.array out_buf = array.array('B', [])

        cdef size_t total_sunk = 0

        while total_sunk < len(in_buf):
            res, sunk = self._encoder.sink(in_buf, offset=total_sunk)

            if res < 0:
                raise RuntimeError('Encoder sink failed.')

            total_sunk += sunk

            # Clear state machine
            for data in self._drain():
                out_buf.extend(data)

        try:
            # Python 3
            return out_buf.tobytes()
        except AttributeError:
            return out_buf.tostring()

    def finish(self):
        """Close encoder and return any remaining data.

        Will throw an exception if fill() or finish()
        is used after this.
        """
        self._check_not_finished()

        cdef array.array out_buf = array.array('B', [])

        while True:
            res = self._encoder.finish()
            if res < 0:
                raise RuntimeError('Encoder finish failed.')

            if self._encoder.is_finished(res):
                self._finished = True
                break

            for data in self._drain():
                out_buf.extend(data)

        try:
            # Python 3
            return out_buf.tobytes()
        except AttributeError:
            return out_buf.tostring()

    @property
    def finished(self):
        """Returns true if the encoder has been closed."""
        return self._finished


cdef _encode_impl(encoder, buf):
    """Encode iterable `buf` into an array of bytes."""
    encoder = Encoder(encoder)
    return encoder.fill(buf) + encoder.finish()


def encode(buf, **kwargs):
    """
    Encode iterable `buf` in to a byte string.

    Keyword arguments:
        window_sz2 (int): Determines how far back in the input can be
            searched for repeated patterns. Defaults to `DEFAULT_WINDOW_SZ2`.
            Allowed values are between. `MIN_WINDOW_SZ2` and `MAX_WINDOW_SZ2`.
        lookahead_sz2 (int): Determines the max length for repeated
            patterns that are found. Defaults to `DEFAULT_LOOKAHEAD_SZ2`.
            Allowed values are between `MIN_LOOKAHEAD_SZ2` and the
            value set for `window_sz2`.

    Returns:
        str or bytes: A byte string of encoded contents.
            str is used for Python 2 and bytes for Python 3.

    Raises:
        ValueError: If `window_sz2` or `lookahead_sz2` are outside their
            defined ranges.
        TypeError: If `window_sz2`, `lookahead_sz2` are not valid numbers and
            if `buf` is not a valid iterable.
        RuntimeError: Thrown if internal polling or sinking of the
            encoder/decoder fails.
    """
    return _encode_impl(Writer(**kwargs), buf)


def decode(buf, **kwargs):
    """
    Decode iterable `buf` in to a byte string.

    Keyword arguments:
        input_buffer_size (int): How large an input buffer to use for the decoder.
            This impacts how much work the decoder can do in a single step,
            a larger buffer will use more memory.
        window_sz2 (int): Determines how far back in the input can be
            searched for repeated patterns. Defaults to `DEFAULT_WINDOW_SZ2`.
            Allowed values are between. `MIN_WINDOW_SZ2` and `MAX_WINDOW_SZ2`.
        lookahead_sz2 (int): Determines the max length for repeated
            patterns that are found. Defaults to `DEFAULT_LOOKAHEAD_SZ2`.
            Allowed values are between `MIN_LOOKAHEAD_SZ2` and the
            value set for `window_sz2`.

    Returns:
        str or bytes: A byte string of decoded contents.
            str is used for Python 2 and bytes for Python 3.

    Raises:
        ValueError: If `input_buffer_size`, `window_sz2` or `lookahead_sz2` are
            outside their defined ranges.
        TypeError: If `input_buffer_size`, `window_sz2` or `lookahead_sz2` are
            not valid numbers and if `buf` is not a valid iterable.
        RuntimeError: Thrown if internal polling or sinking of the
            encoder/decoder fails.
    """
    return _encode_impl(Reader(**kwargs), buf)




import array
import unittest

import heatshrink
from heatshrink.core import Encoder, Reader, Writer

from .constants import TEXT
from .utils import TestUtilsMixin, random_string


class InternalEncodersTest(TestUtilsMixin, unittest.TestCase):
    """Tests for the Writer and Reader classes."""

    def test_checks_window_sz2_type(self):
        for cls in (Writer, Reader):
            self.assertRaises(TypeError, cls, window_sz2='a string')
            self.assertRaises(TypeError, cls, window_sz2=lambda x: None)

    def test_checks_window_sz2_within_limits(self):
        for cls in (Writer, Reader):
            self.assertRaises(ValueError, cls, window_sz2=3)
            self.assertRaises(ValueError, cls, window_sz2=16)
            self.assertNotRaises(cls, window_sz2=5)
            self.assertNotRaises(cls, window_sz2=14)

    # TODO: These kind of tests might be redundant
    def test_checks_lookahead_sz2_type(self):
        for cls in (Writer, Reader):
            self.assertRaises(TypeError, cls, lookahead_sz2='a string')
            self.assertRaises(TypeError, cls, lookahead_sz2=lambda x: None)

    def test_checks_lookahead_sz2_within_limits(self):
        for cls in (Writer, Reader):
            self.assertRaises(ValueError, cls, lookahead_sz2=1)
            self.assertRaises(ValueError, cls, lookahead_sz2=16)
            self.assertNotRaises(cls, lookahead_sz2=4)
            self.assertNotRaises(cls, lookahead_sz2=10)


class EncoderTest(TestUtilsMixin, unittest.TestCase):
    """Test encoder state machine."""

    def setUp(self):
        self.reader = Reader()
        self.writer = Writer()
        self.encoders = [Encoder(e) for e in [self.reader, self.writer]]

    def test_fill_accepted_types(self):
        for encoder in self.encoders:
            self.assertNotRaises(encoder.fill, b'abcde')
            self.assertNotRaises(encoder.fill, u'abcde'.encode('utf8'))
            self.assertNotRaises(encoder.fill, bytearray([1, 2, 3]))
            self.assertNotRaises(encoder.fill, array.array('B', [1, 2, 3]))
            self.assertNotRaises(encoder.fill, [1, 2, 3])

            self.assertRaises(TypeError, encoder.fill, memoryview(b'abcde'))
            self.assertRaises(TypeError, encoder.fill, u'abcde')
            # Obvious fail cases
            self.assertRaises(TypeError, encoder.fill, lambda x: x)
            self.assertRaises(TypeError, encoder.fill, True)

    def test_finished_true_after_finish(self):
        for encoder in self.encoders:
            self.assertTrue(not encoder.finished)
            encoder.finish()
            self.assertTrue(encoder.finished)

    def test_operation_after_finish_fails(self):
        for encoder in self.encoders:
            encoder.fill('abcde')
            encoder.finish()
            self.assertRaises(ValueError, encoder.fill, 'abcde')
            self.assertRaises(ValueError, encoder.finish)

    def test_fill_doesnt_flush_small_values(self):
        encoder = Encoder(self.writer)
        # Pass a small value, this wont cause the encoder
        # to actually do anything
        encoded = encoder.fill('abcde')
        self.assertTrue(len(encoded) == 0)
        # This should clear the encoder completely
        encoded = encoder.finish()
        self.assertTrue(len(encoded) > 0)


class EncodeFunctionTest(TestUtilsMixin, unittest.TestCase):
    """Tests for the core.encode function."""

    def setUp(self):
        self.encoded = heatshrink.encode(b'abcde')

    def test_encoded_size(self):
        self.assertEqual(len(self.encoded), 6)

    def test_encoded_bytes(self):
        self.assertEqual(self.encoded, b'\xb0\xd8\xacvK(')

    def test_encode_with_window_sz2(self):
        encoded = heatshrink.encode(b'abcde', window_sz2=8)
        # FIXME: Prove that this setting changes output
        self.assertEqual(encoded, b'\xb0\xd8\xacvK(')

    def test_encode_with_lookahead_sz2(self):
        encoded = heatshrink.encode(b'abcde', lookahead_sz2=3)
        self.assertEqual(encoded, b'\xb0\xd8\xacvK(')

    def test_different_params_yield_different_output(self):
        string = b'A string with stuff in it'
        self.assertNotEqual(heatshrink.encode(string, window_sz2=8),
                            heatshrink.encode(string, window_sz2=11))
        self.assertNotEqual(heatshrink.encode(string, lookahead_sz2=4),
                            heatshrink.encode(string, lookahead_sz2=8))


class DecodeFunctionTest(TestUtilsMixin, unittest.TestCase):
    """Tests for the core.decode function."""

    def test_returns_string(self):
        self.assertIsInstance(heatshrink.decode('abcde'), str)

    def test_decode_with_window_sz2(self):
        decoded = heatshrink.decode(b'\xb0\xd8\xacvK(', window_sz2=11)
        self.assertEqual(decoded, 'abcde')

    def test_decode_with_lookahead_sz2(self):
        decoded = heatshrink.decode('\xb0\xd8\xacvK(', lookahead_sz2=3)
        self.assertEqual(decoded, 'abcde')


class EncoderToDecoderTest(TestUtilsMixin, unittest.TestCase):
    """
    Tests assertion that data passed through the encoder
    and then the decoder with the same parameters will be
    equal to the original data.
    """

    def test_round_trip(self):
        encoded = heatshrink.encode(b'a string')
        self.assertEqual(heatshrink.decode(encoded), 'a string')

    def test_with_a_paragraph(self):
        encoded = heatshrink.encode(TEXT)
        self.assertEqual(heatshrink.decode(encoded), TEXT)

    def test_with_large_strings(self):
        test_sizes = [1000, 10000, 100000]

        for size in test_sizes:
            contents = random_string(size)

            decoded = heatshrink.decode(heatshrink.encode(contents))
            # Check whole file, but don't use assertEqual as it will
            # print all the data
            if decoded != contents:
                msg = ('Decoded and original file contents '
                       'do not match for size: {}')
                self.fail(msg.format(size))



TEXT = """
Overhead the albatross hangs motionless upon the air
And deep beneath the rolling waves in labyrinths of coral caves
The echo of a distant tide
Comes willowing across the sand
And everything is green and submarine

And no one showed us to the land
And no one knows the wheres or whys
But something stirs and something tries
And starts to climb towards the light

Strangers passing in the street
By chance two separate glances meet
And I am you and what I see is me
And do I take you by the hand
And lead you through the land
And help me understand the best I can
And no one calls us to move on
And no one forces down our eyes
No one speaks
And no one tries
No one flies around the sun

...

Echoes - Pink Floyd
"""

# Compression parameters:
#    window_sz2 - 11
#    lookahead_sz2 - 4
COMPRESSED = (
    b'\x85S\xee\xd6[\x95\xa2\xcba\xb2H.\x96\x8b,\x82\xc3l\xb1Xn\x97+}\xce\xe7 '
    b'\xb4Xm\xd6{\x9c\x82\xdbo\xbaZm\xf6\xebe\x94\x04\n\xebp\xb7\xdb\x80\x8dm7'
    b'(U\x06\xddd\x90Y,\xb6[\x84\x82\xc5e\xb7Yl7K@\x0c%\xca\xdfl\xb6Zm\xd6y'
    b'\x05\xde\xc3v\xb2\xdc\xe4\x16\x9bt\x82\xd9a\xb1^nV\x9bu\xd2\xd1s\x90['
    b'\xec\xd2\x0b\x1d\xbe\xe5a\xb6H,`8p\xaa\xa0\x17\x15\x96\xc7h\xb7\x80\xb9'
    b'\xd8d\x16KM\xce\xe9a\xb7]$\x17KM\x92\xcb\n\xa1\xdb\xed\xa0zW{M\xb2\xd9o'
    b'\xbb\x82i\xd8la\x1e\x80\xc0w;\r\xba\xc8\x0fIe\x0b2\xbc\xdd-\x00<v\x9b'
    b'\x9c\x82\xcfr\xb2\xd9m\xd2\x00\x1b)\x05\xce\xebb\xb6\xd8A\xc4\xac\xb0'
    b'\xa0&Kp/\x95\xb8\r\x8a\xd1o\xbb\xd9l\x92\x0b\xa8"\x15\xbc\x11\xd2\xd8'
    b'\x08\xee\x04\rk\xb7[\xee\xe0\xc2\xb7{E\x96\xe4\x0f\xa5o\xb9\x00I^np\xaa'
    b'\x15\xd6\xe9 \xb9\x84y\x03R\xdc\xee\x96\x9b\x95\xcc\x19\xd4\x04\xe2\xe9r'
    b'\xb4\x86q\x02Y\x86\x11\\\xae\x80\xc4\x96;e\xa6\xdbb\x00\x82\xbb\xd8nV@V'
    b'[e\xa6\xcfh\xbaB\xa1U;\xa5\xc8UJ\xca\x08\x85p\xb0\xdc\xeeaz\x82\xa4\xb7;'
    b'\xa0a\x15\xd2\x15B\xbc\xc8,b\xeeV;,\x82\xe9w\xb7\xc8.v[\x85\x86\xe5a'
    b'\xbaYd\x16p\xaf+\x18J\x95\xb4\x08\xcc\x1a\x8e\x93 \xb0\xdbd\x17\x9b}\xd4"'
    b'\xd2\xefh\xb0\xdd\x00D\xaev[(\x85\x9d\xb6\xca3ko\x01"\xbaXmvP&K\x15\xe4'
    b'\x19P\x16\x88=\x96\xd8=\xc6\x03\x07t\xb4\\\xad\xf7[8\xde\xa8\x8aE\xa2'
    b'\xcbl\xb8\x029H.\xb6\xeb U\x10\xcc\x11\r-\x8a\xcbs\x06s\xb1\xd8m\xc2uv;'
    b'\r\xb2\xd8Cd\x1e\xe9m\xb7\xdd\xac\xa0(@=vk}\xc8.N\xc9o\xbb\xdb\xa4\x16'
    b'\xfb\xad\xcaAe\xbc\x89iS\x86Y\xae\x16[\r\xacN\xd4\x0bXS\x14\x07\x9a\xcdl'
    b'\x00\xc2\x90XB\x80\x82\x01\xaew[t*\x15.\x97K\xa1P\xaa)\x1c\x90\r\x94'
    b'\xb6AP\xb4\xdb\xadr\n5\xb2\xdfy\xb2B\x80'
)



import io
from threading import RLock
from __builtin__ import open as builtin_open

import core

_READ_BUFFER_SIZE = io.DEFAULT_BUFFER_SIZE


class _DecompressReader(io.RawIOBase):
    """Adapts the decompressor API to a RawIOBase reader API.

    This class is similar to the one found in the internal python
    decompression modules. See github for more details:
    https://github.com/python/cpython/blob/3.6/Lib/_compression.py#L33
    """
    def __init__(self, fp, reader_factory, **reader_args):
        self._fp = fp
        self._eof = False
        # Position in file (decompressed)
        self._pos = 0
        # Decompressed data
        self._buf = b''
        self._buf_offset = 0

        # Set to size of decompressed stream once it is known
        self._size = -1

        self._reader_factory = reader_factory
        self._reader_args = reader_args

        self._decoder = self._new_decoder()

    def _new_decoder(self):
        """Create a new decoder using the reader factory and args."""
        reader = self._reader_factory(**self._reader_args)
        return core.Encoder(reader)

    def close(self):
        self._decoder = None
        # Don't close self._fp directly because we don't own it.
        return super(_DecompressReader, self).close()

    def readable(self):
        return True

    def seekable(self):
        try:
            # Python 3
            return self._fp.seekable()
        except AttributeError:
            # Python 2 file object
            # TODO: Check that this is actually robust.
            # TODO: What happens if .seek() is implemented
            # TODO: but not .tell() or .truncate()
            return hasattr(self._fp, 'seek')

    def readinto(self, b):
        buf = self.read(len(b))
        b[:len(buf)] = buf
        return len(buf)

    def _refill(self):
        """Refill internal decompress buffer with file data.

        Throws a EOFError when all data has been read and
        the decoder has been finalized.
        """
        # Concecutive calls should be ignored
        if self._decoder.finished:
            raise EOFError

        self._buf_offset = 0

        raw_chunk = self._fp.read(_READ_BUFFER_SIZE)
        if raw_chunk:
            self._buf = self._decoder.fill(raw_chunk)
        else:
            # Finalize internal decoder.
            self._buf = self._decoder.finish()
            raise EOFError

    def read(self, size=-1):
        if size < 0:
            return self.readall()

        if not size or self._eof:
            return b''

        if self._buf_offset >= len(self._buf):
            try:
                self._refill()
            except EOFError:
                self._eof = True
                self._size = self._pos

        # TODO: Clean up
        data = self._buf[
            self._buf_offset:
            self._buf_offset + size
        ]
        self._buf_offset += size
        self._pos += len(data)

        return data

    def _rewind(self):
        """Rewind the file to the beginning of the data stream."""
        self._fp.seek(0)
        self._eof = False
        self._pos = 0
        self._buf = b''
        self._buf_offset = 0
        # Restart the decoder from the beginning
        self._decoder = self._new_decoder()

    def seek(self, offset, whence=io.SEEK_SET):
        # Recalculate offset as an absolute file position.
        if whence == io.SEEK_SET:
            pass
        elif whence == io.SEEK_CUR:
            offset += self._pos
        elif whence == io.SEEK_END:
            if self._size < 0:
                # Finish reading the file
                while self.read(io.DEFAULT_BUFFER_SIZE):
                    pass
            offset += self._size
        else:
            raise ValueError('Invalid value for whence: {}'.format(whence))

        # Make it so that offset is the number of bytes to skip forward.
        if offset < self._pos:
            self._rewind()
        else:
            offset -= self._pos

        # Read and discard data until we reach the desired position
        while offset > 0:
            data = self.read(min(io.DEFAULT_BUFFER_SIZE, offset))
            if not data:
                break
            offset -= len(data)

        return self._pos

    def tell(self):
        return self._pos


_MODE_CLOSED = 0
_MODE_READ = 1
_MODE_WRITE = 2


class EncodedFile(io.BufferedIOBase):
    def __init__(self, filename, mode='rb', **compress_options):
        """Open a heatshrink LZSS encoded file.

        If filename is a str, bytes or unicode object, it gives the
        name of the file to be opened. Otherwise, it should be a file-like
        object, which will be used to read or write the compressed data.

        mode can be 'rb for reading (default) or 'wb' for (over)writing.
        'r' and 'w' will be converted to to 'rb' and 'wb' respectively.
        """
        self._lock = RLock()
        self._fp = None
        # Should the file be closed by us?
        self._close_fp = False
        self._mode = _MODE_CLOSED

        if mode in ('', 'r', 'rb'):
            mode = 'rb'
            self._mode = _MODE_READ
        elif mode in ('w', 'wb'):
            mode = 'wb'
            self._mode = _MODE_WRITE
        else:
            raise ValueError("Invalid mode: '{!r}'".format(mode))

        if isinstance(filename, (str, bytes, unicode)):
            self._fp = builtin_open(filename, mode)
            # We opened the file, we need to close it
            self._close_fp = True
        elif hasattr(filename, 'read') or hasattr(filename, 'write'):
            # Implements the file protocol
            self._fp = filename
        else:
            msg = 'filename must be an str, bytes or a file-like object'
            raise TypeError(msg)

        if self._mode == _MODE_READ:
            raw = _DecompressReader(self._fp, core.Reader, **compress_options)
            self._buffer = io.BufferedReader(raw)
        else:
            writer = core.Writer(**compress_options)
            self._encoder = core.Encoder(writer)
            # File seek position
            self._pos = 0

        # The file name. Defaults to None
        self.name = getattr(self._fp, 'name', None)

    def seekable(self):
        """Return whether the file supports seeking."""
        return self.readable() and self._buffer.seekable()

    def readable(self):
        """Return whether the file was opened for reading."""
        return self._mode == _MODE_READ

    def writable(self):
        """Return whether the file was opened for writing."""
        return self._mode == _MODE_WRITE

    def _check_not_closed(self):
        """Throws a ValueError if the file has been closed."""
        if self.closed:
            raise ValueError('I/O operation on closed file')

    def _check_can_read(self):
        """Throws an io.UnsupportedOperation if the file can not be read."""
        if not self.readable():
            raise io.UnsupportedOperation('File not open for reading')

    def _check_can_write(self):
        """Throws an io.UnsupportedOperation if the file can not be written."""
        if not self.writable():
            raise io.UnsupportedOperation('File not open for writing')

    def _check_can_seek(self):
        """Throws an io.UnsupportedOperation if the file can not be seeeked."""
        if not self.readable():
            msg = 'Seeking is only supported on files open for reading'
            raise io.UnsupportedOperation(msg)

        if not self.seekable():
            msg = 'The underlying file object does not support seeking.'
            raise io.UnsupportedOperation(msg)

    def close(self):
        """Flush and close the file.

        May be called more than once without error. Once the file is
        closed, any other operation on it will raise a ValueError.
        """
        with self._lock:
            # Flush and finish the decoder.
            if self._mode == _MODE_READ:
                self._buffer.close()
            elif self._mode == _MODE_WRITE:
                self._fp.write(self._encoder.finish())
                self._encoder = None

            try:
                # Actually close the internal file pointer.
                if self._close_fp:
                    self._fp.close()
            finally:
                self._fp = None
                self._close_fp = False
                self._mode = _MODE_CLOSED

    @property
    def closed(self):
        """True if this file is closed."""
        return self._mode == _MODE_CLOSED

    def fileno(self):
        """Return the file descriptor for tthe underlying file."""
        self._check_not_closed()
        return self._fp.fileno()

    def peek(self, n=0):
        """Return buffered data without advancing the file position.

        Always returns at least one byte of data, unless at EOF.
        The exact number of bytes returned is unspecified.
        """
        with self._lock:
            self._check_can_read()
            return self._buffer.peek(n)

    def read(self, size=-1):
        """Read up to size uncompressed bytes from the file.

        If size is negative or omitted, read until EOF is reached.
        Returns b'' if the file is already at EOF.
        """
        self._check_can_read()
        with self._lock:
            return self._buffer.read(size)

    def read1(self, size=-1):
        """Read up to size uncompressed bytes, while trying to avoid
        making multiple reads from the underlying stream. Reads up to a
        buffers worth of data if size is negative.

        Returns b'' if the file is at EOF.
        """
        with self._lock:
            self._check_can_read()
            if size < 0:
                size = io.DEFAULT_BUFFER_SIZE
            return self._buffer.read1(size)

    def readinto(self, b):
        """Read bytes into b.

        Returns the number of bytes read (0 for EOF).
        """
        with self._lock:
            self._check_can_read()
            return self._buffer.readinto(b)

    def readline(self, size=-1):
        """Read a line of uncompressed bytes from the file.

        The terminating newline (if present) is retained. If size is
        non-negative, no more than size bytes will be read (in which
        case the line may be incomplete). Returns b'' if already at EOF.
        """
        with self._lock:
            self._check_can_read()
            return self._buffer.readline(size)

    def readlines(self, size=-1):
        """Read a list of lines of uncompressed bytes from the file.

        size can be specified to control the number of lines read; no
        further lines will be read once the total size of the lines read
        so far equals or exceeds size.
        """
        if not isinstance(size, int):
            if not hasattr(size, '__index__'):
                raise TypeError('Integer argument expected')
            size = size.__index__

        with self._lock:
            self._check_can_read()
            return self._buffer.readlines(size)

    def write(self, data):
        """Write a byte string to the file.

        Returns the number of uncompressed bytes written, which is
        always len(data). Note that due to buffering, the file on disk
        may not reflect the data written until close() is called.
        """
        with self._lock:
            self._check_can_write()
            compressed = self._encoder.fill(data)
            self._fp.write(compressed)
            self._pos += len(data)
            return len(data)

    def writelines(self, seq):
        """Write a sequence of byte strings to the file.

        Returns the number of uncompressed bytes written.
        seq can be any iterable yielding byte strings.

        Line separators are not added between the written byte strings.
        """
        with self._lock:
            self._check_can_write()

            return super(EncodedFile, self).writelines(seq)

    def seek(self, offset, whence=io.SEEK_SET):
        """Change the file position.

        The new position is specified by offset, relative to the
        position indicated by whence. Values for whence are:

            0: start of stream (default); offset must not be negative
            1: current stream position
            2: end of stream; offset must not be positive

        Returns the new file position.

        Note that seeking is emulated, so depending on the parameters,
        this operation may be extremely slow.
        """
        with self._lock:
            self._check_can_seek()
            return self._buffer.seek(offset, whence)

    def tell(self):
        """Return the current file position."""
        with self._lock:
            if self._mode == _MODE_READ:
                return self._buffer.tell()
            return self._pos


def open(filename, mode='rb', **kwargs):
    """Open LZSS compressed file in binary mode.

    The filename argument can be an actual filename (a str, bytes or
    file-like object), or an existing file object to read from or write
    to.

    The mode argument can be "rb", "wb". "r" and "w" are converted to
    "rb" and "rb" respectively.

    This function is equivalent to the EncodedFile constructor:
    EncodedFile(filename, mode, **compress_options).
    """
    return EncodedFile(filename, mode, **kwargs)
